@@ -3,18 +3,21 @@ import os
 import sys
 import threading
 import traceback
-from io import open
 import jaydebeapi
 import jpype
 import re
 import time
-from multiprocessing import Process, Lock
-from multiprocessing.managers import BaseManager
 import setproctitle
 import shlex
 import copy
-from cli_helpers.tabular_output import TabularOutputFormatter, preprocessors
 import click
+import configparser
+import wget
+import hashlib
+from urllib.error import URLError, HTTPError
+from multiprocessing import Process, Lock
+from multiprocessing.managers import BaseManager
+from cli_helpers.tabular_output import TabularOutputFormatter, preprocessors
 from prompt_toolkit.shortcuts import PromptSession
 
 from .sqlexecute import SQLExecute
@@ -138,8 +141,10 @@ class SQLCli(object):
         self.db_saved_conn = {}                   # 数据库Session备s份
         self.SQLMappingHandler = SQLMapping()     # 函数句柄，处理SQLMapping信息
         self.SQLExecuteHandler = SQLExecute()     # 函数句柄，具体来执行SQL
-        self.SQLOptions = SQLOptions()             # 程序运行各种参数
+        self.SQLOptions = SQLOptions()            # 程序运行中各种参数
         self.KafkaHandler = KafkaWrapper()        # 函数句柄，处理Kafka的消息
+        self.SpoolFileHandler = None              # 当前Spool文件句柄
+        self.AppOptions = None                    # 应用程序的配置参数
         self.prompt_app = None                    # PromptKit控制台
         self.db_conn = None                       # 当前应用的数据库连接句柄
         if WorkerName is None:
@@ -305,7 +310,7 @@ class SQLCli(object):
             hidden=False
         )
 
-        # 查看各种信息
+        # 查看各种JOB信息
         register_special_command(
             self.show_job,
             command="showjob",
@@ -313,6 +318,14 @@ class SQLCli(object):
             hidden=False
         )
 
+        # 将SQL信息Spool到指定的文件中
+        register_special_command(
+            self.spool,
+            command="spool",
+            description="spool output to a file",
+            hidden=False
+        )
+
         # 执行特殊的命令
         register_special_command(
             self.execute_internal_command,
@@ -374,6 +387,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Please wait all background process complete.")
 
     # 加载JDBC驱动文件
@@ -448,6 +462,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             'Driver loaded.'
         )
 
@@ -460,6 +475,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             'Mapping file loaded.'
         )
 
@@ -481,6 +497,7 @@ class SQLCli(object):
                         None,
                         None,
                         None,
+                        None,
                         "All Job Completed")
                     return
                 else:
@@ -503,6 +520,7 @@ class SQLCli(object):
                                 None,
                                 None,
                                 None,
+                                None,
                                 "Job " + str(m_Job_ID) + " Completed")
                             return
                     else:
@@ -514,6 +532,7 @@ class SQLCli(object):
                         None,
                         None,
                         None,
+                        None,
                         "No job to wait.")
                     return
                 else:
@@ -540,6 +559,7 @@ class SQLCli(object):
                     None,
                     None,
                     None,
+                    None,
                     "No Jobs")
                 return
 
@@ -560,6 +580,7 @@ class SQLCli(object):
                 None,
                 m_Result,
                 ["JOB#", "ScriptBaseName", "Status", "Started", "End", "Finished", "LoopCount"],
+                None,
                 "Total " + str(self.m_Max_JobID) + " Jobs.")
             return
 
@@ -591,6 +612,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             m_Result)
         return
 
@@ -790,6 +812,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             str(len(m_JobLists)) + " Jobs Started.")
 
     # 提交到后台执行SQL
@@ -857,6 +880,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             "Jobs Submitted. " +
             m_Script_FileName + " Will execute " +
             str(m_Execute_Copies) + " Copies. Loop " + str(m_Script_LoopCount) + " times."
@@ -873,6 +897,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "No Jobs")
             return
 
@@ -900,6 +925,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Total " + str(m_nCount) + " Jobs will shutdown.")
             return
 
@@ -934,12 +960,14 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " will shutdown.")
         else:
             yield (
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " does not exist or does not need shutdown.")
 
     # 强制关闭正在运行的进程
@@ -953,6 +981,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "No Jobs")
             return
 
@@ -976,6 +1005,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Total " + str(m_nCount) + " Jobs will close.")
             return
 
@@ -1005,12 +1035,14 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " aborted.")
         else:
             yield (
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " does not exist or does not need abort.")
 
     # 平和关闭正在运行的进程
@@ -1024,6 +1056,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "No Jobs")
             return
 
@@ -1052,6 +1085,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Total " + str(m_nCount) + " Jobs will close.")
             return
 
@@ -1086,12 +1120,14 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " will close.")
         else:
             yield (
                 None,
                 None,
                 None,
+                None,
                 "Job " + str(m_Job_ID) + " does not exist or does not need close.")
 
     # 连接数据库
@@ -1127,6 +1163,18 @@ class SQLCli(object):
                 connect_parameters[2] + ':' + connect_parameters[3] + ':' + \
                 connect_parameters[4] + '://' + connect_parameters[5] + ':' + \
                 connect_parameters[6] + ':/' + connect_parameters[7]
+        elif len(connect_parameters) == 7:
+            # 数据库连接参数, 但是没有指定driver_type
+            self.db_username = connect_parameters[0]
+            self.db_password = connect_parameters[1]
+            self.db_type = connect_parameters[3]
+            self.db_driver_type = "tcp"
+            self.db_host = connect_parameters[4]
+            self.db_port = connect_parameters[5]
+            self.db_service_name = connect_parameters[6]
+            self.db_url = \
+                connect_parameters[2] + ':' + connect_parameters[3] + ':' + \
+                '://' + connect_parameters[4] + ':' + connect_parameters[5] + ':/' + connect_parameters[6]
         elif len(connect_parameters) == 2:
             # 用户只指定了用户名和口令， 认为用户和上次保留一直的连接字符串信息
             self.db_username = connect_parameters[0]
@@ -1146,12 +1194,22 @@ class SQLCli(object):
                             connect_parameters[0] + ':' + connect_parameters[1] + ':' +\
                             connect_parameters[2] + '://' + connect_parameters[3] + ':' + \
                             connect_parameters[4] + ':/' + connect_parameters[5]
+                    elif len(connect_parameters) == 5:
+                        self.db_type = connect_parameters[1]
+                        self.db_driver_type = 'tcp'
+                        self.db_host = connect_parameters[2]
+                        self.db_port = connect_parameters[3]
+                        self.db_service_name = connect_parameters[4]
+                        self.db_url = \
+                            connect_parameters[0] + ':' + connect_parameters[1] + ':' + \
+                            '://' + connect_parameters[2] + ':' + connect_parameters[3] + ':/' + connect_parameters[4]
                     else:
-                        print("db_type = [" + str(self.db_type) + "]")
-                        print("db_host = [" + str(self.db_host) + "]")
-                        print("db_port = [" + str(self.db_port) + "]")
-                        print("db_service_name = [" + str(self.db_service_name) + "]")
-                        print("db_url = [" + str(self.db_url) + "]")
+                        if "SQLCLI_DEBUG" in os.environ:
+                            print("db_type = [" + str(self.db_type) + "]")
+                            print("db_host = [" + str(self.db_host) + "]")
+                            print("db_port = [" + str(self.db_port) + "]")
+                            print("db_service_name = [" + str(self.db_service_name) + "]")
+                            print("db_url = [" + str(self.db_url) + "]")
                         raise SQLCliException("Unexpeced env SQLCLI_CONNECTION_URL\n." +
                                               "jdbc:[db type]:[driver type]://[host]:[port]/[service name]")
                 else:
@@ -1178,11 +1236,12 @@ class SQLCli(object):
                                 connect_parameters[2] + '://' + connect_parameters[3] + ':' + \
                                 connect_parameters[4] + ':/' + connect_parameters[5]
                         else:
-                            print("db_type = [" + str(self.db_type) + "]")
-                            print("db_host = [" + str(self.db_host) + "]")
-                            print("db_port = [" + str(self.db_port) + "]")
-                            print("db_service_name = [" + str(self.db_service_name) + "]")
-                            print("db_url = [" + str(self.db_url) + "]")
+                            if "SQLCLI_DEBUG" in os.environ:
+                                print("db_type = [" + str(self.db_type) + "]")
+                                print("db_host = [" + str(self.db_host) + "]")
+                                print("db_port = [" + str(self.db_port) + "]")
+                                print("db_service_name = [" + str(self.db_service_name) + "]")
+                                print("db_url = [" + str(self.db_url) + "]")
                             raise SQLCliException("Unexpeced env SQLCLI_CONNECTION_URL\n." +
                                                   "jdbc:[db type]:[driver type]://[host]:[port]/[service name]")
                     else:
@@ -1204,63 +1263,44 @@ class SQLCli(object):
                 jpype.java.lang.Thread.currentThread().setContextClassLoader(
                     jpype.java.lang.ClassLoader.getSystemClassLoader())
 
-            # 拼接所有的Jar包， jaydebeapi将根据class的名字加载指定的文件
+            # 加载所有的Jar包， jaydebeapi将根据class的名字加载指定的文件
             m_JarList = []
+            m_driverclass = ""
+            m_JDBCURL = ""
+            m_JDBCProp = ""
             for m_Jar_Config in self.jar_file:
-                m_JarList.append(m_Jar_Config["FullName"])
-
-            if self.db_type.upper() == "ORACLE":
-                self.db_conn = jaydebeapi.connect("oracle.jdbc.driver.OracleDriver",
-                                                  "jdbc:oracle:thin:@" +
-                                                  self.db_host + ":" + self.db_port + "/" + self.db_service_name,
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            elif self.db_type.upper() == "LINKOOPDB":
-                self.db_conn = jaydebeapi.connect("com.datapps.linkoopdb.jdbc.JdbcDriver",
-                                                  "jdbc:linkoopdb:tcp://" +
-                                                  self.db_host + ":" + self.db_port + "/" + self.db_service_name +
-                                                  ";query_iterator=1",
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            elif self.db_type.upper() == "MYSQL":
-                self.db_conn = jaydebeapi.connect("com.mysql.cj.jdbc.Driver",
-                                                  "jdbc:mysql://" +
-                                                  self.db_host + ":" + self.db_port + "/" + self.db_service_name,
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            elif self.db_type.upper() == "POSTGRESQL":
-                self.db_conn = jaydebeapi.connect("org.postgresql.Driver",
-                                                  "jdbc:postgresql://" +
-                                                  self.db_host + ":" + self.db_port + "/" + self.db_service_name,
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            elif self.db_type.upper() == "SQLSERVER":
-                self.db_conn = jaydebeapi.connect("com.microsoft.sqlserver.jdbc.SQLServerDriver",
-                                                  "jdbc:sqlserver://" +
-                                                  self.db_host + ":" + self.db_port + ";" +
-                                                  "databasename=" + self.db_service_name,
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            elif self.db_type.upper() == "CLICKHOUSE":
-                self.db_conn = jaydebeapi.connect("ru.yandex.clickhouse.ClickHouseDriver",
-                                                  "jdbc:clickhouse://" +
-                                                  self.db_host + ":" + self.db_port + "/" + self.db_service_name,
-                                                  {'user': self.db_username, 'password': self.db_password,
-                                                   'socket_timeout': "360000000"},
-                                                  m_JarList)
-            elif self.db_type.upper() == "TERADATA":
-                self.db_conn = jaydebeapi.connect("com.teradata.jdbc.TeraDriver",
-                                                  "jdbc:teradata://" +
-                                                  self.db_host +
-                                                  "/CLIENT_CHARSET=UTF8,CHARSET=UTF8," +
-                                                  "TMODE=TERA,LOB_SUPPORT=ON,COLUMN_NAME=ON,MAYBENULL=ON," +
-                                                  "database=" + self.db_service_name,
-                                                  [self.db_username, self.db_password],
-                                                  m_JarList)
-            else:
-                raise SQLCliException("Unknown driver. " + self.db_type.upper() + " Exit ")
+                m_JarList.extend(m_Jar_Config["FullName"])
+            for m_Jar_Config in self.jar_file:
+                if m_Jar_Config["Database"].upper() == self.db_type.upper():
+                    m_driverclass = m_Jar_Config["ClassName"]
+                    m_JDBCURL = m_Jar_Config["JDBCURL"]
+                    m_JDBCProp = m_Jar_Config["JDBCProp"]
+                    break
+            if m_JDBCURL is None:
+                raise SQLCliException("Unknown database [" + self.db_type.upper() + "]. Database Connect Failed. \n" +
+                                      "Maybe you forgot sync jlib files. ")
+            m_JDBCURL = m_JDBCURL.replace("${host}", self.db_host)
+            m_JDBCURL = m_JDBCURL.replace("${port}", self.db_port)
+            m_JDBCURL = m_JDBCURL.replace("${service}", self.db_service_name)
+            if m_driverclass is None:
+                raise SQLCliException("Missed driver [" + self.db_type.upper() + "] in config. "
+                                                                                 "Database Connect Failed. ")
+            m_jaydebeapi_prop = {'user': self.db_username, 'password': self.db_password}
+            if m_JDBCProp is not None:
+                for row in m_JDBCProp.strip().split(','):
+                    props = row.split(':')
+                    if len(props) == 2:
+                        m_PropName = str(props[0]).strip()
+                        m_PropValue = str(props[1]).strip()
+                        m_jaydebeapi_prop[m_PropName] = m_PropValue
+            self.db_conn = jaydebeapi.connect(m_driverclass,
+                                              m_JDBCURL,
+                                              m_jaydebeapi_prop,
+                                              m_JarList)
 
             self.SQLExecuteHandler.set_connection(self.db_conn)
+        except SQLCliException as se:  # Connecting to a database fail.
+            raise se
         except Exception as e:  # Connecting to a database fail.
             if "SQLCLI_DEBUG" in os.environ:
                 print('traceback.print_exc():\n%s' % traceback.print_exc())
@@ -1279,6 +1319,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             'Database connected.'
         )
 
@@ -1294,6 +1335,7 @@ class SQLCli(object):
             None,
             None,
             None,
+            None,
             'Database disconnected.'
         )
 
@@ -1313,6 +1355,7 @@ class SQLCli(object):
                     None,
                     None,
                     None,
+                    None,
                     "No saved sesssions."
                 )
             else:
@@ -1320,6 +1363,7 @@ class SQLCli(object):
                     "Saved Sessions",
                     m_Result,
                     ["Sesssion Name", "User Name", "URL"],
+                    None,
                     "Total " + str(len(m_Result)) + " saved sesssions."
                 )
             return
@@ -1349,9 +1393,9 @@ class SQLCli(object):
             raise SQLCliException(
                 "Wrong argument : " + "Session save/restore [session name]")
         if m_Parameters[0] == 'save':
-            yield None, None, None, "Session saved Successful."
+            yield None, None, None, None, "Session saved Successful."
         if m_Parameters[0] == 'restore':
-            yield None, None, None, "Session restored Successful."
+            yield None, None, None, None, "Session restored Successful."
         return
 
     # 休息一段时间, 如果收到SHUTDOWN或者ABORT符号的时候，立刻终止SLEEP
@@ -1372,22 +1416,54 @@ class SQLCli(object):
             time.sleep(int(arg) % 3)
         except ValueError:
             message = "Parameter must be a number, sleep [seconds]."
-            return [(None, None, None, message)]
-        return [(None, None, None, None)]
+            return [(None, None, None, None, message)]
+        return [(None, None, None, None, None)]
 
     # 从文件中执行SQL
     def execute_from_file(self, arg, **_):
         if not arg:
             message = "Missing required argument, filename."
-            return [(None, None, None, message)]
+            return [(None, None, None, None, message)]
         try:
             with open(os.path.expanduser(arg), encoding="utf-8") as f:
                 query = f.read()
         except IOError as e:
-            return [(None, None, None, str(e))]
+            return [(None, None, None, None, str(e))]
 
         return self.SQLExecuteHandler.run(query, os.path.expanduser(arg))
 
+    # 将当前及随后的输出打印到指定的文件中
+    def spool(self, arg, **_):
+        if not arg:
+            message = "Missing required argument, spool [filename]|spool off."
+            return [(None, None, None, None, message)]
+        parameters = str(arg).split()
+        if parameters[0].strip().upper() == 'OFF':
+            # close spool file
+            if self.SpoolFileHandler is None:
+                message = "not spooling currently"
+                return [(None, None, None, None, message)]
+            else:
+                self.SpoolFileHandler.close()
+                self.SpoolFileHandler = None
+                self.SQLExecuteHandler.spoolfile = None
+                return [(None, None, None, None, None)]
+
+        if self.logfilename is not None:
+            # 如果当前主程序启用了日志，则spool日志的默认输出目录为logfile的目录
+            m_FileName = os.path.join(os.path.dirname(self.logfilename), parameters[0].strip())
+        else:
+            # 如果主程序没有启用日志，则输出为当前目录
+            m_FileName = parameters[0].strip()
+        if self.SpoolFileHandler is None:
+            self.SpoolFileHandler = open(m_FileName, "w", encoding="utf-8")
+            self.SQLExecuteHandler.spoolfile = self.SpoolFileHandler
+        else:
+            self.SpoolFileHandler.close()
+            self.SpoolFileHandler = open(m_FileName, "w", encoding="utf-8")
+            self.SQLExecuteHandler.spoolfile = self.SpoolFileHandler
+        return [(None, None, None, None, None)]
+
     # 设置一些选项
     def set_options(self, arg, **_):
         if arg is None:
@@ -1400,12 +1476,14 @@ class SQLCli(object):
                 "Current Options: ",
                 m_Result,
                 ["Name", "Value", "Comments"],
+                None,
                 ""
             )
         else:
             options_parameters = str(arg).split()
             if len(options_parameters) == 1:
-                raise SQLCliException("Missing required argument. set parameter parameter_value.")
+                # 如果没有设置参数，则补充一个None作为参数的值
+                options_parameters.append("")
 
             # 处理DEBUG选项
             if options_parameters[0].upper() == "DEBUG":
@@ -1425,6 +1503,7 @@ class SQLCli(object):
                     None,
                     None,
                     None,
+                    None,
                     '')
             elif self.SQLOptions.get(options_parameters[0].upper()) is not None:
                 self.SQLOptions.set(options_parameters[0].upper(), options_parameters[1])
@@ -1432,6 +1511,7 @@ class SQLCli(object):
                     None,
                     None,
                     None,
+                    None,
                     '')
             else:
                 raise CommandNotFound
@@ -1442,7 +1522,7 @@ class SQLCli(object):
         matchObj = re.match(r"(.*)kafka(.*)$", arg, re.IGNORECASE | re.DOTALL)
         if matchObj:
             (title, result, headers, status) = self.KafkaHandler.Process_SQLCommand(arg)
-            yield title, result, headers, status
+            yield title, result, headers, None, status
             return
 
         # 创建数据文件, 根据末尾的rows来决定创建的行数
@@ -1462,6 +1542,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 str(m_rows) + ' rows created Successful.')
             return
 
@@ -1480,6 +1561,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 str(m_rows) + ' rows created Successful.')
             return
 
@@ -1496,6 +1578,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 'file converted Successful.')
             return
 
@@ -1515,6 +1598,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 'seed file created Successful.')
             return
 
@@ -1533,6 +1617,7 @@ class SQLCli(object):
                 None,
                 None,
                 None,
+                None,
                 'seed file created Successful.')
             return
 
@@ -1583,7 +1668,7 @@ class SQLCli(object):
             result = self.SQLExecuteHandler.run(text)
             # 输出显示结果
             self.formatter.query = text
-            for title, cur, headers, status in result:
+            for title, cur, headers, columntypes, status in result:
                 # 不控制每行的长度
                 max_width = None
 
@@ -1595,7 +1680,7 @@ class SQLCli(object):
                 #   vertical           分行显示，每行、每列都分行
                 #   csv                csv格式显示
                 formatted = self.format_output(
-                    title, cur, headers,
+                    title, cur, headers, columntypes,
                     self.SQLOptions.get("OUTPUT_FORMAT").lower(),
                     max_width
                 )
@@ -1627,6 +1712,59 @@ class SQLCli(object):
             self.m_Current_RunningSQL = None
             self.m_Current_RunningStarted = None
 
+    # 下载程序所需要的各种Jar包
+    def syncdriver(self):
+        # 加载程序的配置文件
+        self.AppOptions = configparser.ConfigParser()
+        m_conf_filename = os.path.join(os.path.dirname(__file__), "conf", "sqlcli.conf")
+        if os.path.exists(m_conf_filename):
+            self.AppOptions.read(m_conf_filename)
+
+        # 下载运行需要的各种Jar包
+        for row in self.AppOptions.items("driver"):
+            print("Checking driver [" + row[0] + "] ... ")
+            for m_driversection in str(row[1]).split(','):
+                m_driversection = m_driversection.strip()
+                try:
+                    m_driver_filename = self.AppOptions.get(m_driversection, "filename")
+                    m_driver_downloadurl = self.AppOptions.get(m_driversection, "downloadurl")
+                    m_driver_filemd5 = self.AppOptions.get(m_driversection, "md5")
+
+                    m_LocalJarFile = os.path.join(os.path.dirname(__file__), "jlib", m_driver_filename)
+                    if os.path.exists(m_LocalJarFile):
+                        with open(m_LocalJarFile, 'rb') as fp:
+                            data = fp.read()
+                        file_md5 = hashlib.md5(data).hexdigest()
+                        if "SQLCLI_DEBUG" in os.environ:
+                            print("File=[" + m_driver_filename + "], MD5=[" + file_md5 + "]")
+                    else:
+                        if "SQLCLI_DEBUG" in os.environ:
+                            print("File=[" + m_driver_filename + "] does not exist!")
+                        file_md5 = ""
+                    if file_md5 != m_driver_filemd5.strip():
+                        print("Driver [" + m_driversection + "], need upgrade ...")
+                        # 重新下载新的文件到本地
+                        try:
+                            wget.download(m_driver_downloadurl, out=m_LocalJarFile)
+                            print("")
+                        except (URLError, HTTPError):
+                            print('traceback.print_exc():\n%s' % traceback.print_exc())
+                            print('traceback.format_exc():\n%s' % traceback.format_exc())
+                            print("")
+                            print("Driver [" + m_driversection + "] download failed.")
+                            continue
+                        with open(m_LocalJarFile, 'rb') as fp:
+                            data = fp.read()
+                        file_md5 = hashlib.md5(data).hexdigest()
+                        if file_md5 != m_driver_filemd5.strip():
+                            print("Driver [" + m_driversection + "] consistent check failed.")
+                        else:
+                            print("Driver [" + m_driversection + "] is up-to-date.")
+                    else:
+                        print("Driver [" + m_driversection + "] is up-to-date.")
+                except (configparser.NoSectionError, configparser.NoOptionError):
+                    print("Bad driver config [" + m_driversection + "], Skip it ...")
+
     # 主程序
     def run_cli(self):
         # 程序运行的结果
@@ -1635,6 +1773,12 @@ class SQLCli(object):
         # 设置主程序的标题，随后开始运行程序
         setproctitle.setproctitle('SQLCli MAIN ' + " Script:" + str(self.sqlscript))
 
+        # 加载程序的配置文件
+        self.AppOptions = configparser.ConfigParser()
+        m_conf_filename = os.path.join(os.path.dirname(__file__), "conf", "sqlcli.conf")
+        if os.path.exists(m_conf_filename):
+            self.AppOptions.read(m_conf_filename)
+
         # 打开输出日志, 如果打开失败，就直接退出
         try:
             if self.logfilename is not None:
@@ -1650,35 +1794,44 @@ class SQLCli(object):
 
         # 加载已经被隐式包含的数据库驱动，文件放置在SQLCli\jlib下
         m_jlib_directory = os.path.join(os.path.dirname(__file__), "jlib")
-        if os.path.isdir(m_jlib_directory):
-            list_file = os.listdir(m_jlib_directory)
-            for i in range(0, len(list_file)):
-                if list_file[i].endswith(".jar"):
-                    m_JarFullFileName = os.path.join(m_jlib_directory, list_file[i])
-                    m_JarBaseFileName = os.path.basename(m_JarFullFileName)
-                    if self.jar_file is None:
-                        self.jar_file = []
-                    if m_JarBaseFileName.startswith("linkoopdb"):
-                        m_DatabaseType = "linkoopdb"
-                    elif m_JarBaseFileName.startswith("mysql"):
-                        m_DatabaseType = "mysql"
-                    elif m_JarBaseFileName.startswith("postgresql"):
-                        m_DatabaseType = "postgresql"
-                    elif m_JarBaseFileName.startswith("sqljdbc"):
-                        m_DatabaseType = "sqlserver"
-                    elif m_JarBaseFileName.startswith("ojdbc"):
-                        m_DatabaseType = "oracle"
-                    elif m_JarBaseFileName.startswith("terajdbc"):
-                        m_DatabaseType = "teradata"
-                    elif m_JarBaseFileName.startswith("tdgssconfig"):
-                        m_DatabaseType = "teradata-gss"
-                    else:
-                        m_DatabaseType = "UNKNOWN"
-                    m_Jar_Config = {"JarName": m_JarBaseFileName,
-                                    "ClassName": None,
-                                    "FullName": m_JarFullFileName,
-                                    "Database": m_DatabaseType}
-                    self.jar_file.append(m_Jar_Config)
+        if self.jar_file is None:
+            self.jar_file = []
+        if self.AppOptions is not None:
+            for row in self.AppOptions.items("driver"):
+                m_DriverName = None
+                m_JarFullFileName = []
+                m_JDBCURL = None
+                m_JDBCProp = None
+                m_DatabaseType = row[0].strip()
+                for m_driversection in str(row[1]).split(','):
+                    m_driversection = m_driversection.strip()
+                    try:
+                        m_jar_filename = self.AppOptions.get(m_driversection, "filename")
+                        if os.path.exists(os.path.join(m_jlib_directory, m_jar_filename)):
+                            m_JarFullFileName.append(os.path.join(m_jlib_directory, m_jar_filename))
+                            if m_DriverName is None:
+                                try:
+                                    m_DriverName = self.AppOptions.get(m_driversection, "driver")
+                                except (configparser.NoSectionError, configparser.NoOptionError):
+                                    m_DriverName = None
+                            if m_JDBCURL is None:
+                                try:
+                                    m_JDBCURL = self.AppOptions.get(m_driversection, "jdbcurl")
+                                except (configparser.NoSectionError, configparser.NoOptionError):
+                                    m_JDBCURL = None
+                            if m_JDBCProp is None:
+                                try:
+                                    m_JDBCProp = self.AppOptions.get(m_driversection, "jdbcprop")
+                                except (configparser.NoSectionError, configparser.NoOptionError):
+                                    m_JDBCProp = None
+                    except (configparser.NoSectionError, configparser.NoOptionError):
+                        pass
+                m_Jar_Config = {"ClassName": m_DriverName,
+                                "FullName": m_JarFullFileName,
+                                "JDBCURL": m_JDBCURL,
+                                "JDBCProp": m_JDBCProp,
+                                "Database": m_DatabaseType}
+                self.jar_file.append(m_Jar_Config)
 
         # 处理传递的映射文件
         if self.sqlmap is not None:   # 如果传递的参数，有Mapping，以参数为准，先加载参数中的Mapping文件
@@ -1755,11 +1908,18 @@ class SQLCli(object):
             click.echo(output, file=self.logfile)
 
     def echo(self, s, **kwargs):
+        # 输出目的地
+        # 1：  程序日志文件 logfile
+        # 2：  程序的logger，用于在第三方调用时候的Console显示
+        # 3：  当前屏幕控制台
+        # 4：  程序的Spool文件
         if self.logfile:
             click.echo(s, file=self.logfile)
         if self.logger is not None:
             self.logger.info(s)
         click.secho(s, **kwargs, file=self.Console)
+        if self.SpoolFileHandler is not None:
+            print(s, file=self.SpoolFileHandler)
 
     def output(self, output, status=None):
         if output:
@@ -1798,7 +1958,33 @@ class SQLCli(object):
         if status:
             self.echo(status)
 
-    def format_output(self, title, cur, headers, p_format_name, max_width=None):
+    def format_output_csv(self, headers, columntypes, cur):
+        m_csv_delimiter = self.SQLOptions.get("CSV_DELIMITER")
+        m_csv_quotechar = self.SQLOptions.get("CSV_QUOTECHAR")
+        if m_csv_delimiter.find("\\t") != -1:
+            m_csv_delimiter = m_csv_delimiter.replace("\\t", '\t')
+        if m_csv_delimiter.find("\\s") != -1:
+            m_csv_delimiter = m_csv_delimiter.replace("\\s", ' ')
+        if self.SQLOptions.get("CSV_HEADER") == "ON":
+            m_row = ""
+            for m_nPos in range(0, len(headers)):
+                m_row = m_row + str(headers[m_nPos])
+                if m_nPos != len(headers) - 1:
+                    m_row = m_row + m_csv_delimiter
+            yield str(m_row)
+
+        for row in cur:
+            m_row = ""
+            for m_nPos in range(0, len(row)):
+                if columntypes[m_nPos] == str:
+                    m_row = m_row + m_csv_quotechar + str(row[m_nPos]) + m_csv_quotechar
+                else:
+                    m_row = m_row + str(row[m_nPos])
+                if m_nPos != len(row) - 1:
+                    m_row = m_row + m_csv_delimiter
+            yield str(m_row)
+
+    def format_output(self, title, cur, headers, columntypes, p_format_name, max_width=None):
         output = []
 
         output_kwargs = {
@@ -1816,13 +2002,17 @@ class SQLCli(object):
             if max_width is not None:
                 cur = list(cur)
 
-            formatted = self.formatter.format_output(
-                cur,
-                headers,
-                format_name=p_format_name,
-                column_types=None,
-                **output_kwargs
-            )
+            if p_format_name.upper() == 'CSV':
+                formatted = self.format_output_csv(headers, columntypes, cur)
+            else:
+                formatted = self.formatter.format_output(
+                    cur,
+                    headers,
+                    format_name=p_format_name,
+                    column_types=None,
+                    **output_kwargs
+                )
+
             if isinstance(formatted, str):
                 formatted = formatted.splitlines()
             formatted = iter(formatted)
@@ -1844,6 +2034,7 @@ class SQLCli(object):
 @click.option("--sqlmap", type=str, help="SQL Mapping file.")
 @click.option("--nologo", is_flag=True, help="Execute with silent mode.")
 @click.option("--sqlperf", type=str, help="SQL performance Log.")
+@click.option("--syncdriver", is_flag=True, help="Download jdbc jar from file server.")
 def cli(
         version,
         logon,
@@ -1851,12 +2042,26 @@ def cli(
         execute,
         sqlmap,
         nologo,
-        sqlperf
+        sqlperf,
+        syncdriver
 ):
     if version:
         print("Version:", __version__)
         sys.exit(0)
 
+    # 从服务器下下载程序需要的各种jar包
+    if syncdriver:
+        sqlcli = SQLCli(
+            logfilename=logfile,
+            logon=logon,
+            sqlscript=execute,
+            sqlmap=sqlmap,
+            nologo=nologo,
+            sqlperf=sqlperf
+        )
+        sqlcli.syncdriver()
+        sys.exit(0)
+
     sqlcli = SQLCli(
         logfilename=logfile,
         logon=logon,
@@ -1871,4 +2076,10 @@ def cli(
 
 
 if __name__ == "__main__":
-    cli()
+    try:
+        cli()
+    except Exception as ge:
+        if "SQLCLI_DEBUG" in os.environ:
+            print('traceback.print_exc():\n%s' % traceback.print_exc())
+            print('traceback.format_exc():\n%s' % traceback.format_exc())
+        print("Fatal Exception: " + repr(ge))
